@@ -19,7 +19,7 @@
  * 
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02111 USA.
  */
 
 #include <Foundation/Foundation.h>
@@ -29,7 +29,6 @@
 #include "PBViewer.h"
 #include "ClipBook.h"
 #include "Functions.h"
-#include "GNUstep.h"
 
 #define WINHEIGHT 415
 #define VWRHEIGHT 300
@@ -134,7 +133,7 @@ static NSString *nibName = @"ClipBookWindow.gorm";
       RELEASE (label);
      
       iconView = [[PBIconView alloc] initWithFrame: [[iconBox contentView] frame]];
-      [(NSBox *)iconBox setContentView: iconView];
+      [iconBox setContentView: iconView];
       
       fm = [NSFileManager defaultManager];
       
@@ -159,7 +158,6 @@ static NSString *nibName = @"ClipBookWindow.gorm";
         pbDescr = [arr mutableCopy];
       }
             
-            
       ASSIGN (dataTypes, ([NSArray arrayWithObjects: NSStringPboardType,
                                                     NSRTFPboardType,
                                                     NSRTFDPboardType,
@@ -171,7 +169,7 @@ static NSString *nibName = @"ClipBookWindow.gorm";
             
       index = 0;
       isDragTarget = NO;
-      [(NSBox *)viewersBox setContentView: emptyBox];
+      [viewersBox setContentView: emptyBox];
       [self showPbData];
     }
   }
@@ -270,7 +268,7 @@ static NSString *nibName = @"ClipBookWindow.gorm";
   [self checkStoredData];
 
   if ([pbDescr count] == 0) {
-    [(NSBox *)viewersBox setContentView: emptyBox];
+    [viewersBox setContentView: emptyBox];
     [self showPbInfo: nil];
   } else {
     NSDictionary *dict = [pbDescr objectAtIndex: index];
@@ -282,10 +280,10 @@ static NSString *nibName = @"ClipBookWindow.gorm";
       id bpviewer = [viewer viewerForData: data ofType: type];
       
       if (bpviewer) {
-        [(NSBox *)viewersBox setContentView: bpviewer];
+        [viewersBox setContentView: bpviewer];
         [self showPbInfo: dict];
       } else {
-        [(NSBox *)viewersBox setContentView: invalidBox];
+        [viewersBox setContentView: invalidBox];
         [self showPbInfo: nil];
       } 
     }
@@ -488,7 +486,7 @@ static NSString *nibName = @"ClipBookWindow.gorm";
 
 @implementation ClipBookWindow (DraggingDestination)
 
-- (unsigned int)draggingEntered:(id <NSDraggingInfo>)sender
+- (NSDragOperation)draggingEntered:(id <NSDraggingInfo>)sender
 {
   NSPasteboard *pb = [sender draggingPasteboard];
   NSArray *types = [pb types];
@@ -509,7 +507,7 @@ static NSString *nibName = @"ClipBookWindow.gorm";
   return NSDragOperationNone;
 }
 
-- (unsigned int)draggingUpdated:(id <NSDraggingInfo>)sender
+- (NSDragOperation)draggingUpdated:(id <NSDraggingInfo>)sender
 {
 	if (isDragTarget) {
 		return NSDragOperationCopy;
