@@ -19,74 +19,35 @@
  * 
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02111 USA.
  */
 
 #include <Foundation/Foundation.h>
 #include <AppKit/AppKit.h>
 #include <AppKit/AppKit.h>
-  #ifdef GNUSTEP 
+#include "FSNode.h"
+#include "FSNFunctions.h"
 #include "GWFunctions.h"
-  #else
-#include <GWorkspace/GWFunctions.h>
-  #endif
 #include "Fiend.h"
 #include "FiendLeaf.h"
 #include "Dialogs/Dialogs.h"
 #include "GWorkspace.h"
 #include "GNUstep.h"
 
-@implementation FiendWindow
-
-- (id)initWithContentRect:(NSRect)contentRect 
-                styleMask:(unsigned int)styleMask 
-                  backing:(NSBackingStoreType)backingType 
-                    defer:(BOOL)flag
-{
-  self = [super initWithContentRect: contentRect
-                  styleMask: styleMask backing: backingType defer: flag];
-  
-  [self registerForDraggedTypes: [NSArray arrayWithObjects: NSFilenamesPboardType, nil]];  
-  
-  return self;
-}
-
-- (unsigned int)draggingEntered:(id <NSDraggingInfo>)sender
-{
-  return [[self contentView] draggingEntered: sender];
-}
-
-- (unsigned int)draggingUpdated:(id <NSDraggingInfo>)sender
-{
-	return [[self contentView] draggingUpdated: sender];
-}
-
-- (void)draggingExited:(id <NSDraggingInfo>)sender
-{
-	[[self contentView] draggingExited: sender];  
-}
-
-- (BOOL)prepareForDragOperation:(id <NSDraggingInfo>)sender
-{
-	return [[self contentView] prepareForDragOperation: sender];
-}
-
-- (BOOL)performDragOperation:(id <NSDraggingInfo>)sender
-{
-	return [[self contentView] performDragOperation: sender];
-}
-
-- (void)concludeDragOperation:(id <NSDraggingInfo>)sender
-{
-  [[self contentView] concludeDragOperation: sender];
-}
-
-@end
-
 @implementation Fiend
 
 - (void)dealloc
 {
+  NSEnumerator *enumerator = [watchedPaths objectEnumerator];  
+  NSString *wpath;
+            
+  while ((wpath = [enumerator nextObject])) {
+    [gw removeWatcherForPath: wpath];
+  }
+	RELEASE (watchedPaths);
+  
+  [[NSNotificationCenter defaultCenter] removeObserver: self];	
+
   RELEASE (layers);
   RELEASE (namelabel);
   RELEASE (ffButt);
@@ -112,19 +73,23 @@
 	
 	  gw = [GWorkspace gworkspace];
 
-	  myWin = [[FiendWindow alloc] initWithContentRect: NSZeroRect
-					               styleMask: NSBorderlessWindowMask  
-                              backing: NSBackingStoreRetained defer: NO];
+	  myWin = [[NSWindow alloc] initWithContentRect: NSZeroRect
+					                      styleMask: NSBorderlessWindowMask  
+                                    backing: NSBackingStoreBuffered defer: NO];
 
-    if ([myWin setFrameUsingName: @"Fiend"] == NO) {
+    if ([myWin setFrameUsingName: @"fiend_window"] == NO) {
       [myWin setFrame: NSMakeRect(100, 100, 64, 64) display: NO];
     }      
-    [myWin setReleasedWhenClosed: NO]; 
     r = [myWin frame];      
-       
+    r.size = NSMakeSize(64, 64);
+    [myWin setFrame: r display: NO];
+    
+    [myWin setReleasedWhenClosed: NO]; 
+    
     defaults = [NSUserDefaults standardUserDefaults];	
 
     layers = [[NSMutableDictionary alloc] initWithCapacity: 1];
+		watchedPaths = [[NSCountedSet alloc] initWithCapacity: 1];
     
     myPrefs = [defaults dictionaryForKey: @"fiendlayers"];
     if (myPrefs != nil) {
@@ -138,19 +103,31 @@
         
         for (j = 0; j < [paths count]; j++) {
           NSString *path = [paths objectAtIndex: j];
+	        NSString *watched = [path stringByDeletingLastPathComponent];	
           
-          if ([[NSFileManager defaultManager] fileExistsAtPath: path] == YES) { 
+          if ([[NSFileManager defaultManager] fileExistsAtPath: path]) { 
             NSDictionary *dict = [pathsAndRects objectForKey: path];
             int posx = [[dict objectForKey: @"posx"] intValue];
             int posy = [[dict objectForKey: @"posy"] intValue];            
 
-            leaf = [[FiendLeaf alloc] initWithPosX: posx posY: posy
-                                relativeToPoint: r.origin forPath: path 
-                                             inFiend: self  ghostImage: nil];
+            leaf = [[FiendLeaf alloc] initWithPosX: posx 
+                                              posY: posy
+                                   relativeToPoint: r.origin 
+                                           forPath: path 
+                                           inFiend: self  
+                                         layerName: layername
+                                        ghostImage: nil];
             [leaves addObject: leaf];
             RELEASE (leaf);
+            
+	          if ([watchedPaths containsObject: watched] == NO) {
+              [gw addWatcherForPath: watched];
+	          }
+            
+            [watchedPaths addObject: watched];
           }                 
-        }        
+        }
+                
         [layers setObject: leaves forKey: layername];        
       }    
       currentName = [defaults stringForKey: @"fiendcurrentlayer"];
@@ -200,7 +177,19 @@
     leaveshidden = NO;
     isDragTarget = NO;
 		
-		[myWin setContentView: self];		
+		[myWin setContentView: self];	
+
+		[[NSNotificationCenter defaultCenter] 
+               addObserver: self 
+                	selector: @selector(fileSystemDidChange:) 
+                			name: @"GWFileSystemDidChangeNotification"
+                		object: nil];                     
+    
+		[[NSNotificationCenter defaultCenter] 
+               addObserver: self 
+                  selector: @selector(watcherNotification:) 
+                		  name: @"GWFileWatcherFileDidChangeNotification"
+                	  object: nil];
   }
   
   return self;
@@ -216,28 +205,6 @@
   return myWin;
 }
 
-- (id)fiendLeafOfType:(NSString *)type withName:(NSString *)name
-{
-  NSArray *leaves = [layers objectForKey: currentName];
-  int i;
-        
-	if (leaveshidden == YES) {
-		return nil;
-	}
-	
-  for (i = 0; i < [leaves count]; i++) {
-    id leaf = [leaves objectAtIndex: i];
-		NSString *ltype = [leaf myType];
-		NSString *lname = [[leaf myPath] lastPathComponent];
-		
-		if (([ltype isEqual: type]) && ([lname isEqual: name])) {
-			return leaf;
-		}
-	}
-	
-	return nil;
-}
-
 - (NSPoint)positionOfLeaf:(id)aleaf
 {
 	return [aleaf iconPosition];
@@ -273,6 +240,13 @@
   }  
   
   layerName = [dialog getEditFieldText];
+
+  if ([layerName length] == 0) {
+		NSString *msg = NSLocalizedString(@"No name supplied!", @"");
+		NSString *buttstr = NSLocalizedString(@"Continue", @"");		
+    NSRunAlertPanel(nil, msg, buttstr, nil, nil);  
+    return;
+  }
   
   if ([[layers allKeys] containsObject: layerName]) {
 		NSString *msg = NSLocalizedString(@"A layer with this name is already present!", @"");
@@ -325,7 +299,17 @@
   leaves = [layers objectForKey: currentName];  
   for (i = 0; i < [leaves count]; i++) {
     id leaf = [leaves objectAtIndex: i];    
-    [[leaf window] close];
+    NSString *watched = [[[leaf node] path] stringByDeletingLastPathComponent];    
+
+	  if ([watchedPaths containsObject: watched]) {
+		  [watchedPaths removeObject: watched];
+      
+      if ([watchedPaths containsObject: watched] == NO) {
+        [gw removeWatcherForPath: watched];
+      }
+	  }
+    
+    [[leaf window] close];    
   }
 
   [layers removeObjectForKey: currentName];     
@@ -359,7 +343,7 @@
   }  
   
   layerName = [dialog getEditFieldText];
-  if ([layerName isEqualToString: currentName]) {  
+  if ([layerName isEqual: currentName]) {  
     return;
   }
   
@@ -567,8 +551,12 @@
       if ([lfpos containsPoint: location]) {
         if (hlightleaf == nil) {        
           hlightleaf = [[FiendLeaf alloc] initWithPosX: [lfpos posx] 
-                  posY: [lfpos posy] relativeToPoint: r.origin forPath: nil 
-                                       inFiend: self ghostImage: [leaf icon]];
+                                                  posY: [lfpos posy] 
+                                       relativeToPoint: r.origin 
+                                               forPath: nil 
+                                               inFiend: self 
+                                             layerName: nil
+                                            ghostImage: [leaf icon]];
           [[hlightleaf window] display];                           
           [[hlightleaf window] orderBack: self];                                 
         } else {        
@@ -606,6 +594,16 @@
     }
       
     if (newpos == NO) {    
+      NSString *watched = [[[leaf node] path] stringByDeletingLastPathComponent];    
+
+	    if ([watchedPaths containsObject: watched]) {
+		    [watchedPaths removeObject: watched];
+        
+        if ([watchedPaths containsObject: watched] == NO) {
+          [gw removeWatcherForPath: watched];
+        }
+	    }
+    
       [[leaf window] close];
       [leaves removeObject: leaf];
     }
@@ -752,6 +750,235 @@
   }
 }
 
+- (void)removeInvalidLeaf:(FiendLeaf *)leaf
+{
+  NSString *layerName = [leaf layerName];
+  NSMutableArray *leaves = [layers objectForKey: layerName];
+  NSString *watched = [[[leaf node] path] stringByDeletingLastPathComponent];    
+
+  if ([watchedPaths containsObject: watched]) {
+    [watchedPaths removeObject: watched];
+    
+    if ([watchedPaths containsObject: watched] == NO) {
+      [gw removeWatcherForPath: watched];
+    }
+  }
+  
+  [[leaf window] close];
+  [leaves removeObject: leaf]; 
+}
+
+- (void)checkIconsAfterDotsFilesChange
+{
+  NSArray *names = [layers allKeys];
+  int i;
+
+  for (i = 0; i < [names count]; i++) {
+    NSString *lname = [names objectAtIndex: i];
+    NSMutableArray *leaves = [layers objectForKey: lname];
+    int count = [leaves count];
+    BOOL modified = NO;  
+    int j;
+
+    for (j = 0; j < count; j++) {
+      id leaf = [leaves objectAtIndex: j];
+      NSString *leafpath = [[leaf node] path];
+      
+      if ([leafpath rangeOfString: @"."].location != NSNotFound) {
+        [self removeInvalidLeaf: leaf];
+        modified = YES;   
+        count--;
+        j--;
+      }
+    }
+    
+    if (modified && ([lname isEqual: currentName])) {
+      [self orderFrontLeaves];
+      [self findFreePositions];
+    }
+  }
+}
+
+- (void)checkIconsAfterHidingOfPaths:(NSArray *)paths
+{
+  NSArray *names = [layers allKeys];
+  int i;
+
+  for (i = 0; i < [names count]; i++) {
+    NSString *lname = [names objectAtIndex: i];
+    NSMutableArray *leaves = [layers objectForKey: lname];
+    int count = [leaves count];
+    BOOL modified = NO;  
+    int j, m;
+
+    for (j = 0; j < count; j++) {
+      id leaf = [leaves objectAtIndex: j];
+      NSString *leafpath = [[leaf node] path];
+      
+      for (m = 0; m < [paths count]; m++) {
+        NSString *path = [paths objectAtIndex: m]; 
+      
+        if (isSubpathOfPath(path, leafpath) || [path isEqual: leafpath]) {
+          [self removeInvalidLeaf: leaf];          
+          modified = YES;   
+          count--;
+          j--;
+          break;
+        }
+      }
+    }
+    
+    if (modified && ([lname isEqual: currentName])) {
+      [self orderFrontLeaves];
+      [self findFreePositions];
+    }
+  }
+}
+
+- (void)fileSystemDidChange:(NSNotification *)notification
+{
+  CREATE_AUTORELEASE_POOL(arp);
+  NSDictionary *dict = [notification object];
+  NSString *operation = [dict objectForKey: @"operation"];
+  NSString *source = [dict objectForKey: @"source"];
+  NSArray *files = [dict objectForKey: @"files"];
+  
+  if ([operation isEqual: @"GWorkspaceRenameOperation"]) {
+		files = [NSArray arrayWithObject: [source lastPathComponent]];
+    source = [source stringByDeletingLastPathComponent];
+  }	
+
+  if ([operation isEqual: NSWorkspaceMoveOperation] 
+        || [operation isEqual: NSWorkspaceDestroyOperation]
+				|| [operation isEqual: @"GWorkspaceRenameOperation"]
+				|| [operation isEqual: NSWorkspaceRecycleOperation]
+				|| [operation isEqual: @"GWorkspaceRecycleOutOperation"]
+				|| [operation isEqual: @"GWorkspaceEmptyRecyclerOperation"]) {
+    NSMutableArray *paths = [NSMutableArray arrayWithCapacity: 1];
+    NSArray *names = [layers allKeys];    
+    int i;
+    
+    for (i = 0; i < [files count]; i++) {
+      NSString *s = [source stringByAppendingPathComponent: [files objectAtIndex: i]];
+      [paths addObject: s];
+    }
+
+    for (i = 0; i < [names count]; i++) {
+      NSString *lname = [names objectAtIndex: i];
+      NSMutableArray *leaves = [layers objectForKey: lname];
+      int count = [leaves count];
+      BOOL modified = NO;  
+      int j, m;
+
+      for (j = 0; j < count; j++) {
+        id leaf = [leaves objectAtIndex: j];
+        NSString *leafpath = [[leaf node] path];
+
+        for (m = 0; m < [paths count]; m++) {
+          NSString *path = [paths objectAtIndex: m]; 
+
+          if (isSubpathOfPath(path, leafpath) || [path isEqual: leafpath]) {
+            [self removeInvalidLeaf: leaf];          
+            modified = YES;   
+            count--;
+            j--;
+            break;
+          }
+        }
+      }
+      
+      if (modified && ([lname isEqual: currentName])) {
+        [self orderFrontLeaves];
+        [self findFreePositions];
+      }
+    }
+  }
+  
+  RELEASE (arp);
+}
+
+- (void)watcherNotification:(NSNotification *)notification
+{
+  CREATE_AUTORELEASE_POOL(arp);
+	NSDictionary *notifdict = (NSDictionary *)[notification object];
+  NSString *path = [notifdict objectForKey: @"path"];
+	NSString *event = [notifdict objectForKey: @"event"];
+  NSEnumerator *enumerator;
+  NSString *wpath;
+	BOOL contained = NO;
+	
+	if ([event isEqual: @"GWFileCreatedInWatchedDirectory"]) {
+    RELEASE (arp);
+		return;
+	}
+  
+  enumerator = [watchedPaths objectEnumerator];  
+            
+  while ((wpath = [enumerator nextObject])) {
+		if (([wpath isEqual: path]) || (isSubpathOfPath(path, wpath))) {
+			contained = YES;
+			break;
+		}
+  }
+
+  if (contained) {
+    NSArray *names = [layers allKeys];
+    int i;
+
+    for (i = 0; i < [names count]; i++) {
+      NSString *lname = [names objectAtIndex: i];
+      NSMutableArray *leaves = [layers objectForKey: lname];
+      int count = [leaves count];
+      BOOL modified = NO;  
+      int j;
+      
+      if ([event isEqual: @"GWWatchedPathDeleted"]) {
+        for (j = 0; j < count; j++) {
+          id leaf = [leaves objectAtIndex: j];
+          NSString *leafpath = [[leaf node] path];
+
+				  if (isSubpathOfPath(path, leafpath)) {
+					  [self removeInvalidLeaf: leaf];
+            modified = YES; 
+					  count--;
+					  j--;
+				  }
+        }
+
+      } else if ([event isEqual: @"GWFileDeletedInWatchedDirectory"]) {
+			  NSArray *files = [notifdict objectForKey: @"files"];
+      
+        for (j = 0; j < count; j++) {
+          id leaf = [leaves objectAtIndex: j];
+          NSString *leafpath = [[leaf node] path];
+          int m;
+          
+          for (m = 0; m < [files count]; m++) {
+            NSString *fname = [files objectAtIndex: m];
+            NSString *fullPath = [path stringByAppendingPathComponent: fname];
+
+            if ((isSubpathOfPath(fullPath, leafpath))
+															      || ([fullPath isEqual: leafpath])) {
+					    [self removeInvalidLeaf: leaf];
+              modified = YES; 
+					    count--;
+					    j--;
+              break;
+            }
+          }
+        }
+      }
+      
+      if (modified && ([lname isEqual: currentName])) {
+        [self orderFrontLeaves];
+        [self findFreePositions];
+      }
+    }
+  }
+  
+  RELEASE (arp);
+}
+
 - (void)updateDefaults
 {
 	NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];		
@@ -769,7 +996,7 @@
       id leaf = [leaves objectAtIndex: j];
       [dict setObject: [NSString stringWithFormat: @"%i", [leaf posx]] forKey: @"posx"];      
       [dict setObject: [NSString stringWithFormat: @"%i", [leaf posy]] forKey: @"posy"];      
-      [pathsAndRects setObject: dict forKey: [leaf myPath]];
+      [pathsAndRects setObject: dict forKey: [[leaf node] path]];
     }
     
     [prefs setObject: pathsAndRects forKey: name];    
@@ -777,9 +1004,8 @@
 
  	[defaults setObject: prefs forKey: @"fiendlayers"];  
   [defaults setObject: currentName forKey: @"fiendcurrentlayer"];  
-	[defaults synchronize];
   
-  [myWin saveFrameUsingName: @"Fiend"];
+  [myWin saveFrameUsingName: @"fiend_window"];
 }
 
 - (void)drawRect:(NSRect)rect
@@ -798,7 +1024,7 @@
 
 @implementation Fiend (DraggingDestination)
 
-- (unsigned int)draggingEntered:(id <NSDraggingInfo>)sender
+- (NSDragOperation)draggingEntered:(id <NSDraggingInfo>)sender
 {
 	NSPasteboard *pb = [sender draggingPasteboard];
 	
@@ -817,7 +1043,7 @@
   return NSDragOperationNone;
 }
 
-- (unsigned int)draggingUpdated:(id <NSDraggingInfo>)sender
+- (NSDragOperation)draggingUpdated:(id <NSDraggingInfo>)sender
 {
   NSDragOperation sourceDragMask;
 	
@@ -855,6 +1081,7 @@
   NSPasteboard *pb;
 	NSArray *sourcePaths;
   NSString *path;
+  NSString *basepath;
   NSMutableArray *leaves;
   id leaf;
   NSRect r;
@@ -875,10 +1102,16 @@
   leaves = [layers objectForKey: currentName];  
 
   path = [sourcePaths objectAtIndex: 0];
-  
+  basepath = [path stringByDeletingLastPathComponent];
+
+  if ([basepath isEqual: [gw trashPath]]) {
+    isDragTarget = NO;
+    return;
+  }
+    
   for (i = 0; i < [leaves count]; i++) {
     leaf = [leaves objectAtIndex: i];    
-    if ([[leaf myPath] isEqualToString: path] == YES) {
+    if ([[[leaf node] path] isEqual: path] == YES) {
 			NSString *msg = NSLocalizedString(@"This object is already present in this layer!", @"");
 			NSString *buttstr = NSLocalizedString(@"Continue", @"");		
       NSRunAlertPanel(nil, msg, buttstr, nil, nil);
@@ -901,12 +1134,22 @@
   }
   posy--;
                   
-  leaf = [[FiendLeaf alloc] initWithPosX: posx posY: posy
-                            relativeToPoint: r.origin forPath: path 
-                                          inFiend: self  ghostImage: nil];                          
+  leaf = [[FiendLeaf alloc] initWithPosX: posx 
+                                    posY: posy
+                         relativeToPoint: r.origin 
+                                 forPath: path 
+                                 inFiend: self 
+                               layerName: currentName 
+                              ghostImage: nil];                          
   [leaves addObject: leaf];
   RELEASE (leaf);
   
+	if ([watchedPaths containsObject: basepath] == NO) {
+    [gw addWatcherForPath: basepath];
+	}
+  
+  [watchedPaths addObject: basepath];
+  
   leaf = [leaves objectAtIndex: [leaves count] -1];
   [[leaf window] display]; 
   [self findFreePositions];  
