@@ -19,27 +19,22 @@
  * 
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02111 USA.
  */
 
 #include <Foundation/Foundation.h>
 #include <AppKit/AppKit.h>
-  #ifdef GNUSTEP 
-#include "GWLib.h"
-#include "GWFunctions.h"
-#include "GWNotifications.h"
-  #else
-#include <GWorkspace/GWLib.h>
-#include <GWorkspace/GWFunctions.h>
-#include <GWorkspace/GWNotifications.h>
-  #endif
+#include <math.h>
+#include "FSNodeRep.h"
+#include "FSNFunctions.h"
 #include "TShelfIconsView.h"
 #include "TShelfIcon.h"
 #include "TShelfPBIcon.h"
 #include "GWorkspace.h"
 #include "GNUstep.h"
 
-#define CELLS_WIDTH 80
+#define CELLS_WIDTH (80)
+#define EDIT_MARGIN (4)
 
 @interface TShelfIcon (TShelfIconsViewSorting)
 
@@ -95,6 +90,7 @@
 	RELEASE (icons);  
 	RELEASE (watchedPaths);
 	TEST_RELEASE (dragImage);
+  RELEASE (focusedIconLabel);
   [super dealloc];
 }
 
@@ -105,7 +101,7 @@
 	self = [super init];
   
   if (self) {
-    NSArray *hiddenPaths = [GWLib hiddenPaths];
+    NSArray *hiddenPaths = [[FSNodeRep sharedInstance] hiddenPaths];
     int i, j;
 
     fm = [NSFileManager defaultManager];
@@ -119,8 +115,20 @@
 		
     cellsWidth = CELLS_WIDTH;
     		
-		watchedPaths = [[NSMutableArray alloc] initWithCapacity: 1];
-		
+		watchedPaths = [[NSCountedSet alloc] initWithCapacity: 1];
+		    
+    focusedIconLabel = [NSTextField new];
+		[focusedIconLabel setFont: [NSFont systemFontOfSize: 12]];
+		[focusedIconLabel setBezeled: NO];
+		[focusedIconLabel setAlignment: NSCenterTextAlignment];
+    [focusedIconLabel setEditable: NO];
+    [focusedIconLabel setSelectable: NO];
+    [focusedIconLabel setBackgroundColor: [NSColor windowBackgroundColor]];
+	  [focusedIconLabel setTextColor: [NSColor controlTextColor]];
+    [focusedIconLabel setFrame: NSMakeRect(0, 0, 0, 14)];
+    
+    focusedIcon = nil;
+    
 		icons = [[NSMutableArray alloc] initWithCapacity: 1];
         
     iconsType = itype;
@@ -172,13 +180,13 @@
 		    [[NSNotificationCenter defaultCenter] 
                    addObserver: self 
                 	    selector: @selector(fileSystemWillChange:) 
-                			    name: GWFileSystemWillChangeNotification
+                			    name: @"GWFileSystemWillChangeNotification"
                 		    object: nil];
 
 		    [[NSNotificationCenter defaultCenter] 
                    addObserver: self 
                 	    selector: @selector(fileSystemDidChange:) 
-                			    name: GWFileSystemDidChangeNotification
+                			    name: @"GWFileSystemDidChangeNotification"
                 		    object: nil];                     
       } else {
         NSArray *types = [NSArray arrayWithObjects: NSStringPboardType,
@@ -196,7 +204,7 @@
 		  [[NSNotificationCenter defaultCenter] 
                  addObserver: self 
                     selector: @selector(watcherNotification:) 
-                		    name: GWFileWatcherFileDidChangeNotification
+                		    name: @"GWFileWatcherFileDidChangeNotification"
                 	    object: nil];
 	  }
   }
@@ -252,9 +260,9 @@
 	[self resizeWithOldSuperviewSize: [self frame].size];  
 
 	if ([watchedPaths containsObject: watched] == NO) {
-		[watchedPaths addObject: watched];
 		[self setWatcherForPath: watched];
 	}
+  [watchedPaths addObject: watched];
 }
 
 - (TShelfPBIcon *)addPBIconForDataAtPath:(NSString *)dpath 
@@ -281,23 +289,39 @@
 
 - (void)removeIcon:(id)anIcon
 {
-  int index = [anIcon gridindex];	
+  if (anIcon) {
+    id label = [anIcon myLabel];
+    int index = [anIcon gridindex];	
 
-  if (iconsType == FILES_TAB) {
-	  NSString *watched = [[[anIcon paths] objectAtIndex: 0] stringByDeletingLastPathComponent];
+    if (iconsType == FILES_TAB) {
+	    NSString *watched = [[[anIcon paths] objectAtIndex: 0] stringByDeletingLastPathComponent];
 
-	  if ([watchedPaths containsObject: watched]) {
-		  [watchedPaths removeObject: watched];
-		  [self unsetWatcherForPath: watched];
-	  }
-  
-    [[anIcon myLabel] removeFromSuperview];
+	    if ([watchedPaths containsObject: watched]) {
+		    [watchedPaths removeObject: watched];
+        
+        if ([watchedPaths containsObject: watched] == NO) {
+		      [self unsetWatcherForPath: watched];
+        }
+	    }
+      
+      if (label && [[self subviews] containsObject: label]) {
+        [label removeFromSuperview];
+      }
+    }
+    
+    if (focusedIcon == anIcon) {
+      focusedIcon = nil;
+      [self updateFocusedIconLabel];
+    }
+    
+    if ([[self subviews] containsObject: anIcon]) {
+      [anIcon removeFromSuperview];
+    }
+    
+    [icons removeObject: anIcon];
+    gpoints[index].used = 0;
+    [self resizeWithOldSuperviewSize: [self frame].size];  
   }
-  
-  [anIcon removeFromSuperview];
-  [icons removeObject: anIcon];
-  gpoints[index].used = 0;
-  [self resizeWithOldSuperviewSize: [self frame].size];  
 }
 
 - (void)removePBIconsWithData:(NSData *)data ofType:(NSString *)type
@@ -334,7 +358,7 @@
 	iconwidth = [icon frame].size.width;
 	labwidth = [label frame].size.width;
 
-	if(iconwidth > labwidth) {
+	if (iconwidth > labwidth) {
 		labxpos = [icon frame].origin.x + ((iconwidth - labwidth) / 2);
 	} else {
 		labxpos = [icon frame].origin.x - ((labwidth - iconwidth) / 2);
@@ -370,14 +394,63 @@
   }  
 }
 
+- (void)setFocusedIcon:(id)anIcon
+{
+  if (anIcon == nil) {
+    if (focusedIcon) {
+      [self addSubview: [focusedIcon myLabel]];
+      [self setLabelRectOfIcon: focusedIcon];
+    }
+  } 
+
+  focusedIcon = anIcon;  
+  [self updateFocusedIconLabel];
+}
+
+- (void)updateFocusedIconLabel
+{
+  if ([[self subviews] containsObject: focusedIconLabel]) {
+    NSRect rect = [focusedIconLabel frame];
+
+    [focusedIconLabel removeFromSuperview];
+    [self setNeedsDisplayInRect: rect];
+  }
+  
+  if (focusedIcon) {
+    NSRect iconrect = [focusedIcon frame];
+    float centerx = iconrect.origin.x + (iconrect.size.width / 2);  
+    NSTextField *label = [focusedIcon myLabel];
+    NSRect labelrect = [label frame];
+    NSString *name = [focusedIcon shownName];  
+    float fwidth = [[label font] widthOfString: name];
+    float boundswidth = [self bounds].size.width - EDIT_MARGIN;
+    int margin = 8;
+      
+    fwidth += margin;  
+
+    if ((centerx + (fwidth / 2)) >= boundswidth) {
+      centerx -= (centerx + (fwidth / 2) - boundswidth);
+    } else if ((centerx - (fwidth / 2)) < margin) {
+      centerx += fabs(centerx - (fwidth / 2)) + margin;
+    }    
+
+    labelrect.origin.x = centerx - (fwidth / 2);
+    labelrect.size.width = fwidth;
+    labelrect = NSIntegralRect(labelrect);
+
+    [label removeFromSuperview];
+
+    [focusedIconLabel setFrame: labelrect];
+    [focusedIconLabel setStringValue: name];
+    [self addSubview: focusedIconLabel];  
+    [self setNeedsDisplayInRect: labelrect];
+  }  
+}
+
 - (void)sortIcons
 {
   SEL sel = (iconsType == FILES_TAB) ? @selector(iconCompare:) : @selector(pbiconCompare:);
-	NSArray *sortedIcons = [icons sortedArrayUsingSelector: sel];	
-  RETAIN (sortedIcons);
-	[icons removeAllObjects];
-	[icons addObjectsFromArray: sortedIcons];
-  RELEASE (sortedIcons);
+  [icons sortUsingSelector: sel];
 }
 
 - (NSArray *)icons
@@ -395,7 +468,11 @@
   int i;
 
   for (i = 0; i < [icons count]; i++) {
-    [[icons objectAtIndex: i] renewIcon];
+    id icon = [icons objectAtIndex: i];
+    
+    if ([icon respondsToSelector: @selector(renewIcon)]) {
+      [icon renewIcon];
+    }
   }  
 }
 
@@ -420,7 +497,32 @@
 
 - (void)openCurrentSelection:(NSArray *)paths
 {
-  [gw openSelectedPaths: paths newViewer: NO];
+  [gw openSelectedPaths: paths newViewer: YES];
+}
+
+- (void)checkIconsAfterDotsFilesChange
+{
+  if (iconsType == FILES_TAB) {
+    int count = [icons count]; 
+    int i;
+
+	  for (i = 0; i < count; i++) {
+		  TShelfIcon *icon = [icons objectAtIndex: i];
+      NSArray *iconpaths = [icon paths];
+      int j;
+
+	    for (j = 0; j < [iconpaths count]; j++) {
+        NSString *op = [iconpaths objectAtIndex: j];
+
+        if ([op rangeOfString: @"."].location != NSNotFound) {
+          [self removeIcon: icon];
+          count--;
+          i--;
+          break;
+        }
+      }
+	  }
+  }
 }
 
 - (void)checkIconsAfterHidingOfPaths:(NSArray *)hpaths
@@ -442,7 +544,7 @@
 	      for (m = 0; m < [hpaths count]; m++) {
           NSString *fp = [hpaths objectAtIndex: m]; 
 
-          if (subPathOfPath(fp, op) || [fp isEqualToString: op]) {  
+          if (isSubpathOfPath(fp, op) || [fp isEqual: op]) {  
             [self removeIcon: icon];
             count--;
             i--;
@@ -465,19 +567,19 @@
 
 - (void)fileSystemWillChange:(NSNotification *)notification
 {
-  NSDictionary *dict = [notification userInfo];
+  CREATE_AUTORELEASE_POOL(arp);
+  NSDictionary *dict = [notification object];
   NSString *operation = [dict objectForKey: @"operation"];
 	NSString *source = [dict objectForKey: @"source"];	  
 	NSArray *files = [dict objectForKey: @"files"];	 
 
   if ([operation isEqual: NSWorkspaceMoveOperation] 
         || [operation isEqual: NSWorkspaceDestroyOperation]
-				|| [operation isEqual: GWorkspaceRenameOperation]
+				|| [operation isEqual: @"GWorkspaceRenameOperation"]
 				|| [operation isEqual: NSWorkspaceRecycleOperation]
-				|| [operation isEqual: GWorkspaceRecycleOutOperation]
-				|| [operation isEqual: GWorkspaceEmptyRecyclerOperation]) {
-    
-    NSMutableArray *paths = [NSMutableArray arrayWithCapacity: 1];
+				|| [operation isEqual: @"GWorkspaceRecycleOutOperation"]
+				|| [operation isEqual: @"GWorkspaceEmptyRecyclerOperation"]) {
+    NSMutableArray *paths = [NSMutableArray array];
     NSArray *iconpaths;
     int i, j, m;
 
@@ -505,52 +607,35 @@
       }
 	  }
   }
+
+  RELEASE (arp);
 }
 
 - (void)fileSystemDidChange:(NSNotification *)notification
 {
-  NSDictionary *dict;
-  NSString *operation, *source, *destination;
-  NSArray *files;
-  NSMutableArray *paths;
-  TShelfIcon *icon;
-  NSArray *iconpaths;
-  int count;
-	int i, j, m;
+  CREATE_AUTORELEASE_POOL(arp);
+  NSDictionary *dict = [notification object];
+  NSString *operation = [dict objectForKey: @"operation"];
+  NSString *source = [dict objectForKey: @"source"];
+  NSArray *files = [dict objectForKey: @"files"];
   
-  dict = [notification userInfo];
-  operation = [dict objectForKey: @"operation"];
-  source = [dict objectForKey: @"source"];
-  destination = [dict objectForKey: @"destination"];
-  files = [dict objectForKey: @"files"];
-		                    
-  if ([operation isEqual: GWorkspaceRenameOperation]) {      
-    for (i = 0; i < [icons count]; i++) {
-      icon = [icons objectAtIndex: i];      
-      if ([icon isSinglePath] == YES) {      
-        if ([[[icon paths] objectAtIndex: 0] isEqualToString: source]) {     
-          [icon setPaths: [NSArray arrayWithObject: destination]];
-          [icon setNeedsDisplay: YES];
-					[self resizeWithOldSuperviewSize: [self frame].size];  
-          break;
-        }
-      }          
-    }        
-  }  
-
-  if ([operation isEqual: GWorkspaceRenameOperation]) {
+  if ([operation isEqual: @"GWorkspaceRenameOperation"]) {
 		files = [NSArray arrayWithObject: [source lastPathComponent]];
     source = [source stringByDeletingLastPathComponent];
   }	
 		                    
   if ([operation isEqual: NSWorkspaceMoveOperation] 
         || [operation isEqual: NSWorkspaceDestroyOperation]
-				|| [operation isEqual: GWorkspaceRenameOperation]
+				|| [operation isEqual: @"GWorkspaceRenameOperation"]
 				|| [operation isEqual: NSWorkspaceRecycleOperation]
-				|| [operation isEqual: GWorkspaceRecycleOutOperation]
-				|| [operation isEqual: GWorkspaceEmptyRecyclerOperation]) {
+				|| [operation isEqual: @"GWorkspaceRecycleOutOperation"]
+				|| [operation isEqual: @"GWorkspaceEmptyRecyclerOperation"]) {
+    NSMutableArray *paths = [NSMutableArray arrayWithCapacity: 1];
+    TShelfIcon *icon;
+    NSArray *iconpaths;
+    int count;
+	  int i, j, m;
 
-    paths = [NSMutableArray arrayWithCapacity: 1];
     for (i = 0; i < [files count]; i++) {
       NSString *s = [source stringByAppendingPathComponent: [files objectAtIndex: i]];
       [paths addObject: s];
@@ -579,7 +664,6 @@
           if (deleted) {
             break;
           } 
-
         }
 
         if (deleted) {
@@ -588,20 +672,25 @@
       }
 	  }
   }
+  
+  RELEASE (arp);
 }
 
 - (void)watcherNotification:(NSNotification *)notification
 {
+  CREATE_AUTORELEASE_POOL(arp);
 	NSDictionary *notifdict = (NSDictionary *)[notification object];
   NSString *path = [notifdict objectForKey: @"path"];
 	NSString *event = [notifdict objectForKey: @"event"];
-	BOOL contained = NO;
+	NSEnumerator *enumerator;
+  NSString *wpath;
+  BOOL contained = NO;
 	int i;
   
   if (iconsType == DATA_TAB) {
     int count = [icons count];
     
-    if (event == GWFileDeletedInWatchedDirectory) { 
+    if ([event isEqual: @"GWFileDeletedInWatchedDirectory"]) { 
       NSArray *files = [notifdict objectForKey: @"files"];
     
       for (i = 0; i < count; i++) {
@@ -613,7 +702,7 @@
           NSString *fname = [files objectAtIndex: j];
           NSString *fullPath = [path stringByAppendingPathComponent: fname];
 
-          if ([fullPath isEqualToString: dataPath]) {
+          if ([fullPath isEqual: dataPath]) {
             [self removeIcon: icon];
             count--;
             i--;
@@ -622,17 +711,19 @@
       }
     }
   } else {
-	  if (event == GWFileCreatedInWatchedDirectory) {
+	  if ([event isEqual: @"GWFileCreatedInWatchedDirectory"]) {
+      RELEASE (arp);
 		  return;
 	  }
-
-	  for (i = 0; i < [watchedPaths count]; i++) {
-		  NSString *wpath = [watchedPaths objectAtIndex: i];
-		  if (([wpath isEqualToString: path]) || (subPathOfPath(path, wpath))) {
+    
+    enumerator = [watchedPaths objectEnumerator]; 
+    
+    while ((wpath = [enumerator nextObject])) {
+		  if (([wpath isEqual: path]) || (isSubpathOfPath(path, wpath))) {
 			  contained = YES;
 			  break;
 		  }
-	  }
+    }
 
     if (contained) {
 		  id icon;
@@ -640,22 +731,24 @@
 		  NSString *ipath;
 		  int count = [icons count];
 
-		  if (event == GWWatchedDirectoryDeleted) {		
+		  if ([event isEqual: @"GWWatchedPathDeleted"]) {		
 			  for (i = 0; i < count; i++) {
 				  icon = [icons objectAtIndex: i];
 				  ipaths = [icon paths];
 				  ipath = [ipaths objectAtIndex: 0];
 
-				  if (subPathOfPath(path, ipath)) {
+				  if (isSubpathOfPath(path, ipath)) {
 					  [self removeIcon: icon];
 					  count--;
 					  i--;
 				  }
 			  }
+        
+        RELEASE (arp);
 			  return;
 		  }		
 
-		  if (event == GWFileDeletedInWatchedDirectory) { 
+		  if ([event isEqual: @"GWFileDeletedInWatchedDirectory"]) { 
 			  NSArray *files = [notifdict objectForKey: @"files"];
 
 			  for (i = 0; i < count; i++) {
@@ -671,8 +764,8 @@
 						  NSString *fname = [files objectAtIndex: j];
 						  NSString *fullPath = [path stringByAppendingPathComponent: fname];
 
-						  if ((subPathOfPath(fullPath, ipath))
-															  || ([ipath isEqualToString: fullPath])) {
+						  if ((isSubpathOfPath(fullPath, ipath))
+															  || ([ipath isEqual: fullPath])) {
 							  [self removeIcon: icon];
 							  count--;
 							  i--;
@@ -681,7 +774,6 @@
 					  }
 
 				  } else {
-
 					  for (j = 0; j < [files count]; j++) {
 						  NSString *fname = [files objectAtIndex: j];
 						  NSString *fullPath = [path stringByAppendingPathComponent: fname];
@@ -693,7 +785,7 @@
 						  }
 
 						  ipath = [ipaths objectAtIndex: 0];
-						  if (subPathOfPath(fullPath, ipath)) {
+						  if (isSubpathOfPath(fullPath, ipath)) {
 							  [self removeIcon: icon];
 							  count--;
 							  i--;
@@ -703,7 +795,7 @@
 						  for (m = 0; m < [ipaths count]; m++) {
 							  ipath = [ipaths objectAtIndex: m];
 
-							  if ([ipath isEqualToString: fullPath]) {
+							  if ([ipath isEqual: fullPath]) {
 								  NSMutableArray *newpaths;
 
 								  if ([ipaths count] == 1) {
@@ -728,34 +820,38 @@
 		  }
 	  }
   }
+
+  RELEASE (arp);
 }
 
 - (void)setWatchers
 {
-	int i;
-	
-  for (i = 0; i < [watchedPaths count]; i++) {
-    [self setWatcherForPath: [watchedPaths objectAtIndex: i]];  
+  NSEnumerator *enumerator = [watchedPaths objectEnumerator]; 
+  NSString *wpath;
+
+	while ((wpath = [enumerator nextObject])) {
+    [self setWatcherForPath: wpath]; 
   }
 }
 
 - (void)setWatcherForPath:(NSString *)path
 {
-	[GWLib addWatcherForPath: path];
+	[gw addWatcherForPath: path];
 }
 
 - (void)unsetWatchers
 {
-	int i;
-	
-  for (i = 0; i < [watchedPaths count]; i++) {
-    [self unsetWatcherForPath: [watchedPaths objectAtIndex: i]];  
+  NSEnumerator *enumerator = [watchedPaths objectEnumerator]; 
+  NSString *wpath;
+
+	while ((wpath = [enumerator nextObject])) {
+    [self unsetWatcherForPath: wpath]; 
   }
 }
 
 - (void)unsetWatcherForPath:(NSString *)path
 {
-	[GWLib removeWatcherForPath: path];
+	[gw removeWatcherForPath: path];
 }
 
 - (void)makePositions
@@ -865,7 +961,7 @@
 		NSRect r = NSMakeRect(p.x, p.y, 64, 52);
 
 		[icon setPosition: p];
-		[icon setFrame: r];
+		[icon setFrame: NSIntegralRect(r)];
 		gpoints[index].used = 1;
     
     if (iconsType == FILES_TAB) {
@@ -1029,13 +1125,15 @@
 
 @implementation TShelfIconsView(DraggingDestination)
 
-- (unsigned int)draggingEntered:(id <NSDraggingInfo>)sender
+- (NSDragOperation)draggingEntered:(id <NSDraggingInfo>)sender
 {
   NSPasteboard *pb = [sender draggingPasteboard];
   NSDragOperation sourceDragMask = [sender draggingSourceOperationMask];
   BOOL found = YES;
   gridpoint *gpoint;
   
+  DESTROY (dragImage);
+  
   if (iconsType == FILES_TAB) {
 	  if ((sourceDragMask == NSDragOperationCopy) 
 												  || (sourceDragMask == NSDragOperationLink)) {
@@ -1044,7 +1142,22 @@
   }
   
   if (iconsType == FILES_TAB) {  
-    if ([[pb types] indexOfObject: NSFilenamesPboardType] == NSNotFound) {
+    if ([[pb types] indexOfObject: NSFilenamesPboardType] != NSNotFound) {
+      NSArray *sourcePaths = [pb propertyListForType: NSFilenamesPboardType]; 
+      
+      if ([sourcePaths count]) {
+        NSString *basePath = [sourcePaths objectAtIndex: 0];
+      
+        basePath = [basePath stringByDeletingLastPathComponent];
+        
+        if ([basePath isEqual: [gw trashPath]]) {
+          found = NO;
+        }
+      
+      } else {
+        found = NO;
+      }
+    } else {
       found = NO;
     }
   } else {
@@ -1084,7 +1197,7 @@
   return NSDragOperationNone;
 }
 
-- (unsigned int)draggingUpdated:(id <NSDraggingInfo>)sender
+- (NSDragOperation)draggingUpdated:(id <NSDraggingInfo>)sender
 {
   NSDragOperation sourceDragMask = [sender draggingSourceOperationMask];
 	NSPoint p = [sender draggedImageLocation];
@@ -1097,6 +1210,10 @@
   if (iconsType == FILES_TAB) {
 	  if ((sourceDragMask == NSDragOperationCopy) 
 												  || (sourceDragMask == NSDragOperationLink)) {
+      if (dragImage) {
+        DESTROY (dragImage);
+        [self setNeedsDisplayInRect: dragRect]; 
+      }                   
 		  return NSDragOperationNone;
 	  }	
   }
@@ -1119,9 +1236,7 @@
 			[self setNeedsDisplayInRect: dragRect];
 
 		} else {
-			if (dragImage != nil) {
-				DESTROY (dragImage);
-			}
+      DESTROY (dragImage);
 			return NSDragOperationNone;
 		}
 	}
