@@ -5,7 +5,7 @@
  * Author: Enrico Sersale <enrico@imago.ro>
  * Date: March 2004
  *
- * This file is part of the GNUstep Operation application
+ * This file is part of the GNUstep GWorkspace application
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -19,7 +19,7 @@
  * 
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02111 USA.
  */
 
 #include <Foundation/Foundation.h>
@@ -29,19 +29,22 @@
 #include "Functions.h"
 #include "GNUstep.h"
 
+#define PROGR_STEPS (100.0)
+static BOOL stopped = NO;
+static BOOL paused = NO;
+
 static NSString *nibName = @"FileOperationWin";
 
 @implementation FileOpInfo
 
 - (void)dealloc
 {
-	[nc removeObserver: self];
-
   RELEASE (operationDict);
   RELEASE (type);
   TEST_RELEASE (source);
   TEST_RELEASE (destination);
   TEST_RELEASE (files);
+  TEST_RELEASE (dupfiles);
   TEST_RELEASE (notifNames);
   TEST_RELEASE (win);
   TEST_RELEASE (progInd);
@@ -101,8 +104,8 @@ static NSString *nibName = @"FileOperationWin";
       }
 
       RETAIN (progInd);
-      r = [[(NSBox *)progBox contentView] frame];
-      progView = [[ProgressView alloc] initWithFrame: r refreshInterval: 0.05];
+      r = [[progBox contentView] frame];
+      progView = [[OpProgressView alloc] initWithFrame: r refreshInterval: 0.05];
 
       [fromLabel setStringValue: NSLocalizedString(@"From:", @"")];
       [toLabel setStringValue: NSLocalizedString(@"To:", @"")];
@@ -119,7 +122,50 @@ static NSString *nibName = @"FileOperationWin";
     ASSIGN (type, tp);
     ASSIGN (source, src);
     ASSIGN (destination, dst);
-		files = [fls mutableCopy];	
+    ASSIGN (files, fls);
+    
+    dupfiles = [NSMutableArray new];
+    
+    if ([type isEqual: @"NSWorkspaceDuplicateOperation"]) {    
+      NSString *copystr = NSLocalizedString(@"_copy", @"");
+      unsigned i;
+      
+      for (i = 0; i < [files count]; i++) {
+        NSDictionary *fdict = [files objectAtIndex: i];
+        NSString *fname = [fdict objectForKey: @"name"]; 
+        NSString *newname = [NSString stringWithString: fname];
+        NSString *ext = [newname pathExtension]; 
+        NSString *base = [newname stringByDeletingPathExtension];        
+        NSString *ntmp;
+	      NSString *destpath;        
+        int count = 1;
+	      
+	      while (1) {
+          if (count == 1) {
+            ntmp = [NSString stringWithFormat: @"%@%@", base, copystr];
+            if ([ext length]) {
+              ntmp = [ntmp stringByAppendingPathExtension: ext];
+            }
+          } else {
+            ntmp = [NSString stringWithFormat: @"%@%@%i", base, copystr, count];
+            if ([ext length]) {
+              ntmp = [ntmp stringByAppendingPathExtension: ext];
+            }
+          }
+
+		      destpath = [destination stringByAppendingPathComponent: ntmp];
+
+		      if ([fm fileExistsAtPath: destpath] == NO) {
+            newname = ntmp;
+			      break;
+          } else {
+            count++;
+          }
+	      }
+        
+        [dupfiles addObject: newname];
+      }
+    }
     
     operationDict = [NSMutableDictionary new];
     [operationDict setObject: type forKey: @"operation"]; 
@@ -144,30 +190,52 @@ static NSString *nibName = @"FileOperationWin";
   if (confirm) {    
 	  NSString *title;
 	  NSString *msg, *msg1, *msg2;
+    NSString *items;
 
+    if ([files count] > 1) {
+      items = [NSString stringWithFormat: @"%i %@", [files count], NSLocalizedString(@"items", @"")];
+    } else {
+      items = NSLocalizedString(@"one item", @"");
+    }
+    
 	  if ([type isEqual: @"NSWorkspaceMoveOperation"]) {
 		  title = NSLocalizedString(@"Move", @"");
-		  msg1 = NSLocalizedString(@"Move from: ", @"");
+      msg1 = [NSString stringWithFormat: @"%@ %@ %@: ", 
+                                            NSLocalizedString(@"Move", @""), 
+                                            items, 
+                                            NSLocalizedString(@"from", @"")];
 		  msg2 = NSLocalizedString(@"\nto: ", @"");
 		  msg = [NSString stringWithFormat: @"%@%@%@%@?", msg1, source, msg2, destination];
-	  } else if ([type isEqual: @"NSWorkspaceCopyOperation"]) {
+    } else if ([type isEqual: @"NSWorkspaceCopyOperation"]) {
 		  title = NSLocalizedString(@"Copy", @"");
-		  msg1 = NSLocalizedString(@"Copy from: ", @"");
+      msg1 = [NSString stringWithFormat: @"%@ %@ %@: ", 
+                                            NSLocalizedString(@"Copy", @""), 
+                                            items, 
+                                            NSLocalizedString(@"from", @"")];
 		  msg2 = NSLocalizedString(@"\nto: ", @"");
 		  msg = [NSString stringWithFormat: @"%@%@%@%@?", msg1, source, msg2, destination];
 	  } else if ([type isEqual: @"NSWorkspaceLinkOperation"]) {
 		  title = NSLocalizedString(@"Link", @"");
-		  msg1 = NSLocalizedString(@"Link ", @"");
+      msg1 = [NSString stringWithFormat: @"%@ %@ %@: ", 
+                                            NSLocalizedString(@"Link", @""), 
+                                            items, 
+                                            NSLocalizedString(@"from", @"")];
 		  msg2 = NSLocalizedString(@"\nto: ", @"");
 		  msg = [NSString stringWithFormat: @"%@%@%@%@?", msg1, source, msg2, destination];
 	  } else if ([type isEqual: @"NSWorkspaceRecycleOperation"]) {
 		  title = NSLocalizedString(@"Recycler", @"");
-		  msg1 = NSLocalizedString(@"Move from: ", @"");
+      msg1 = [NSString stringWithFormat: @"%@ %@ %@: ", 
+                                            NSLocalizedString(@"Move", @""), 
+                                            items, 
+                                            NSLocalizedString(@"from", @"")];
 		  msg2 = NSLocalizedString(@"\nto the Recycler", @"");
 		  msg = [NSString stringWithFormat: @"%@%@%@?", msg1, source, msg2];
 	  } else if ([type isEqual: @"GWorkspaceRecycleOutOperation"]) {
 		  title = NSLocalizedString(@"Recycler", @"");
-		  msg1 = NSLocalizedString(@"Move from the Recycler ", @"");
+      msg1 = [NSString stringWithFormat: @"%@ %@ %@ ", 
+                                            NSLocalizedString(@"Move", @""), 
+                                            items, 
+                                            NSLocalizedString(@"from the Recycler", @"")];
 		  msg2 = NSLocalizedString(@"\nto: ", @"");
 		  msg = [NSString stringWithFormat: @"%@%@%@?", msg1, msg2, destination];
 	  } else if ([type isEqual: @"GWorkspaceEmptyRecyclerOperation"]) {
@@ -180,7 +248,7 @@ static NSString *nibName = @"FileOperationWin";
 		  title = NSLocalizedString(@"Duplicate", @"");
 		  msg = NSLocalizedString(@"Duplicate the selected objects?", @"");
 	  }
-
+        
     if (NSRunAlertPanel(title, msg, 
                         NSLocalizedString(@"OK", @""), 
 				                NSLocalizedString(@"Cancel", @""), 
@@ -199,7 +267,8 @@ static NSString *nibName = @"FileOperationWin";
 				                                      sendPort: port[1]];
   [execconn setRootObject: self];
   [execconn setDelegate: self];
-
+  RETAIN (execconn);
+  
   [nc addObserver: self
          selector: @selector(connectionDidDie:)
              name: NSConnectionDidDieNotification
@@ -221,57 +290,44 @@ static NSString *nibName = @"FileOperationWin";
       [self endOperation];
     }
   NS_ENDHANDLER
-
-  [nc addObserver: self
-         selector: @selector(threadWillExit:)
-             name: NSThreadWillExitNotification
-           object: nil];    
-}
-
-- (void)threadWillExit:(NSNotification *)notification
-{
-  NSLog(@"operation thread will exit");
 }
 
 - (int)requestUserConfirmationWithMessage:(NSString *)message 
                                     title:(NSString *)title
-{
+{  
   return NSRunAlertPanel(NSLocalizedString(title, @""),
 												 NSLocalizedString(message, @""),
-											   NSLocalizedString(@"OK", @""), 
+											   NSLocalizedString(@"Ok", @""), 
 												 NSLocalizedString(@"Cancel", @""), 
                          nil);       
 }
 
 - (int)showErrorAlertWithMessage:(NSString *)message
-{
+{  
   return NSRunAlertPanel(nil, 
                          NSLocalizedString(message, @""), 
-												 NSLocalizedString(@"Continue", @""), 
+												 NSLocalizedString(@"Ok", @""), 
                          nil, 
                          nil);
 }
 
 - (IBAction)pause:(id)sender
 {
-  if (executor) {  
-	  if ([executor isPaused] == NO) {
-		  [pauseButt setTitle: NSLocalizedString(@"Continue", @"")];
-		  [stopButt setEnabled: NO];	
-      [executor Pause];
-	  } else {
-		  [pauseButt setTitle: NSLocalizedString(@"Pause", @"")];
-		  [stopButt setEnabled: YES];	
-		  [executor performOperation];
-	  }
-  }
+	if (paused == NO) {
+		[pauseButt setTitle: NSLocalizedString(@"Continue", @"")];
+		[stopButt setEnabled: NO];	
+    paused = YES;
+	} else {
+		[pauseButt setTitle: NSLocalizedString(@"Pause", @"")];
+		[stopButt setEnabled: YES];	
+    paused = NO;
+		[executor performOperation];
+	}
 }
 
 - (IBAction)stop:(id)sender
 {
-  if (executor) {
-	  [executor Stop];
-  }
+  stopped = YES;   
 }
 
 - (void)showProgressWin
@@ -280,42 +336,42 @@ static NSString *nibName = @"FileOperationWin";
     if ([type isEqual: @"NSWorkspaceMoveOperation"]) {
       [win setTitle: NSLocalizedString(@"Move", @"")];
       [fromLabel setStringValue: NSLocalizedString(@"From:", @"")];
-      [fromField setStringValue: relativePathFittingInContainer(fromField, source)];
+      [fromField setStringValue: relativePathFittingInField(fromField, source)];
       [toLabel setStringValue: NSLocalizedString(@"To:", @"")];
-      [toField setStringValue: relativePathFittingInContainer(fromField, destination)];
+      [toField setStringValue: relativePathFittingInField(fromField, destination)];
     
     } else if ([type isEqual: @"NSWorkspaceCopyOperation"]) {
       [win setTitle: NSLocalizedString(@"Copy", @"")];
       [fromLabel setStringValue: NSLocalizedString(@"From:", @"")];
-      [fromField setStringValue: relativePathFittingInContainer(fromField, source)];
+      [fromField setStringValue: relativePathFittingInField(fromField, source)];
       [toLabel setStringValue: NSLocalizedString(@"To:", @"")];
-      [toField setStringValue: relativePathFittingInContainer(fromField, destination)];
+      [toField setStringValue: relativePathFittingInField(fromField, destination)];
     
     } else if ([type isEqual: @"NSWorkspaceLinkOperation"]) {
       [win setTitle: NSLocalizedString(@"Link", @"")];
       [fromLabel setStringValue: NSLocalizedString(@"From:", @"")];
-      [fromField setStringValue: relativePathFittingInContainer(fromField, source)];
+      [fromField setStringValue: relativePathFittingInField(fromField, source)];
       [toLabel setStringValue: NSLocalizedString(@"To:", @"")];
-      [toField setStringValue: relativePathFittingInContainer(fromField, destination)];
+      [toField setStringValue: relativePathFittingInField(fromField, destination)];
     
     } else if ([type isEqual: @"NSWorkspaceDuplicateOperation"]) {
       [win setTitle: NSLocalizedString(@"Duplicate", @"")];
       [fromLabel setStringValue: NSLocalizedString(@"In:", @"")];
-      [fromField setStringValue: relativePathFittingInContainer(fromField, destination)];
+      [fromField setStringValue: relativePathFittingInField(fromField, destination)];
       [toLabel setStringValue: @""];
       [toField setStringValue: @""];
     
     } else if ([type isEqual: @"NSWorkspaceDestroyOperation"]) {
       [win setTitle: NSLocalizedString(@"Destroy", @"")];
       [fromLabel setStringValue: NSLocalizedString(@"In:", @"")];
-      [fromField setStringValue: relativePathFittingInContainer(fromField, destination)];
+      [fromField setStringValue: relativePathFittingInField(fromField, destination)];
       [toLabel setStringValue: @""];
       [toField setStringValue: @""];
     
     } else if ([type isEqual: @"NSWorkspaceRecycleOperation"]) {
       [win setTitle: NSLocalizedString(@"Move", @"")];
       [fromLabel setStringValue: NSLocalizedString(@"From:", @"")];
-      [fromField setStringValue: relativePathFittingInContainer(fromField, source)];
+      [fromField setStringValue: relativePathFittingInField(fromField, source)];
       [toLabel setStringValue: NSLocalizedString(@"To:", @"")];
       [toField setStringValue: NSLocalizedString(@"the Recycler", @"")];
         
@@ -324,7 +380,7 @@ static NSString *nibName = @"FileOperationWin";
       [fromLabel setStringValue: NSLocalizedString(@"From:", @"")];
       [fromField setStringValue: NSLocalizedString(@"the Recycler", @"")];
       [toLabel setStringValue: NSLocalizedString(@"To:", @"")];
-      [toField setStringValue: relativePathFittingInContainer(fromField, destination)];
+      [toField setStringValue: relativePathFittingInField(fromField, destination)];
                             
     } else if ([type isEqual: @"GWorkspaceEmptyRecyclerOperation"]) {
       [win setTitle: NSLocalizedString(@"Destroy", @"")];
@@ -334,46 +390,45 @@ static NSString *nibName = @"FileOperationWin";
       [toField setStringValue: @""];    
     }
     
-    [(NSBox *)progBox setContentView: progView];
+    [progBox setContentView: progView];
     [progView start];
   }
   
-  [win makeKeyAndOrderFront: nil];
+  [win orderFront: nil];
   showwin = YES;
 }
 
 - (void)setNumFiles:(int)n
 {
-  filescount = n;
   [progView stop];  
-  [(NSBox *)progBox setContentView: progInd];
+  [progBox setContentView: progInd];
   [progInd setMinValue: 0.0];
-  [progInd setMaxValue: filescount];
+  [progInd setMaxValue: n];
   [progInd setDoubleValue: 0.0];
-  
   [executor performOperation]; 
 }
 
-- (void)updateProgressIndicator
+- (void)setProgIndicatorValue:(int)n
 {
-  [progInd incrementBy: 1.0];
+  [progInd setDoubleValue: n];
 }
 
 - (void)endOperation
 {
   if (showwin) {
-    if ([(NSBox *)progBox contentView] == progView) {
+    if ([progBox contentView] == progView) {
       [progView stop];  
     }
     [win saveFrameUsingName: @"fopinfo"];
     [win close];
   }
+
+  [nc removeObserver: self];
   
   if (executor) {
     [nc removeObserver: self
 	                name: NSConnectionDidDieNotification 
                 object: execconn];
-    [executor exitThread];
     DESTROY (executor);
     DESTROY (execconn);
   }
@@ -383,60 +438,35 @@ static NSString *nibName = @"FileOperationWin";
 
 - (void)sendWillChangeNotification
 {
-  NSString *destpath;
-	NSMutableDictionary *dict;
+  CREATE_AUTORELEASE_POOL(arp);
+	NSMutableDictionary *dict = [NSMutableDictionary dictionary];	
   int i;
     
   notifNames = [NSMutableArray new];
   
-  if ([type isEqual: @"NSWorkspaceDuplicateOperation"]) { 
-    NSString *copystr = NSLocalizedString(@"copy", @"");
-    NSString *ofstr = NSLocalizedString(@"_of_", @"");
-        
-    for (i = 0; i < [files count]; i++) {
-      NSDictionary *dict = [files objectAtIndex: i];
-      NSString *name = [dict objectForKey: @"name"]; 
-      NSString *ntmp;
-      int count = 1;
-          
-			while(1) {
-        if (count == 1) {
-          ntmp = [NSString stringWithFormat: @"%@%@%@", copystr, ofstr, name];
-        } else {
-          ntmp = [NSString stringWithFormat: @"%@%i%@%@", copystr, count, ofstr, name];
-        }
-        
-				destpath = [destination stringByAppendingPathComponent: ntmp];        
-				if ([fm fileExistsAtPath: destpath] == NO) {  
-          [notifNames addObject: ntmp];
-					break;
-        } else {
-          count++;
-        }
-			}
-    }
-  } else {
-    for (i = 0; i < [files count]; i++) {
-      NSDictionary *dict = [files objectAtIndex: i];
-      NSString *name = [dict objectForKey: @"name"]; 
-
-      [notifNames addObject: name];
-    }
+  for (i = 0; i < [files count]; i++) {
+    NSDictionary *fdict = [files objectAtIndex: i];
+    NSString *name = [fdict objectForKey: @"name"]; 
+    [notifNames addObject: name];
   }
   
-	dict = [NSMutableDictionary dictionary];		
 	[dict setObject: type forKey: @"operation"];	
   [dict setObject: source forKey: @"source"];	
   [dict setObject: destination forKey: @"destination"];	
   [dict setObject: notifNames forKey: @"files"];	
 
+  [nc postNotificationName: @"GWFileSystemWillChangeNotification"
+	 								  object: dict];
+
 	[dnc postNotificationName: @"GWFileSystemWillChangeNotification"
 	 								   object: nil 
                    userInfo: dict];
+  RELEASE (arp);
 }
 
 - (void)sendDidChangeNotification
 {
+  CREATE_AUTORELEASE_POOL(arp);
   NSMutableDictionary *notifObj = [NSMutableDictionary dictionary];		
 
 	[notifObj setObject: type forKey: @"operation"];	
@@ -456,9 +486,13 @@ static NSString *nibName = @"FileOperationWin";
   
   opdone = YES;			
 
+  [nc postNotificationName: @"GWFileSystemDidChangeNotification"
+	 								  object: notifObj];
+
 	[dnc postNotificationName: @"GWFileSystemDidChangeNotification"
 	 						       object: nil 
                    userInfo: notifObj];  
+  RELEASE (arp);
 }
 
 - (void)registerExecutor:(id)anObject
@@ -496,7 +530,7 @@ static NSString *nibName = @"FileOperationWin";
 			msg = @"Some items have the same name;\ndo you want to replace them?";
 			title = @"Recycle";
 		}
-  
+        
     result = NSRunAlertPanel(NSLocalizedString(title, @""),
 														 NSLocalizedString(msg, @""),
 														 NSLocalizedString(@"OK", @""), 
@@ -515,8 +549,10 @@ static NSString *nibName = @"FileOperationWin";
     [self showProgressWin];
   }
 
-  filescount = -1;
-  [self sendWillChangeNotification];  
+  [self sendWillChangeNotification]; 
+  
+  stopped = NO;
+  paused = NO;   
   [executor calculateNumFiles];
 }
 
@@ -572,6 +608,11 @@ static NSString *nibName = @"FileOperationWin";
   return files;
 }
 
+- (NSArray *)dupfiles
+{
+  return dupfiles;
+}
+
 - (int)ref
 {
   return ref;
@@ -602,13 +643,11 @@ static NSString *nibName = @"FileOperationWin";
 
 + (void)setPorts:(NSArray *)thePorts
 {
-  NSAutoreleasePool *pool;
+  CREATE_AUTORELEASE_POOL(pool);
   NSPort *port[2];
   NSConnection *conn;
   FileOpExecutor *executor;
-               
-  pool = [[NSAutoreleasePool alloc] init];
-               
+                              
   port[0] = [thePorts objectAtIndex: 0];             
   port[1] = [thePorts objectAtIndex: 1];             
 
@@ -619,8 +658,7 @@ static NSString *nibName = @"FileOperationWin";
   [executor setFileop: thePorts];
   [(id)[conn rootProxy] registerExecutor: executor];
   RELEASE (executor);
-                              
-  [[NSRunLoop currentRunLoop] run];
+  
   RELEASE (pool);
 }
 
@@ -640,8 +678,6 @@ static NSString *nibName = @"FileOperationWin";
   
   if (self) {
     fm = [NSFileManager defaultManager];
-		stopped = NO;
-		paused = NO;  
 		samename = NO;
     onlyolder = NO;
   }
@@ -670,7 +706,6 @@ static NSString *nibName = @"FileOperationWin";
 {
   NSDictionary *opDict = [NSUnarchiver unarchiveObjectWithData: opinfo];
   id dictEntry;
-  int i;
 
   dictEntry = [opDict objectForKey: @"operation"];
   if (dictEntry) {
@@ -690,9 +725,7 @@ static NSString *nibName = @"FileOperationWin";
   files = [NSMutableArray new];
   dictEntry = [opDict objectForKey: @"files"];
   if (dictEntry) {
-    for (i = 0; i < [dictEntry count]; i++) {
-      [files addObject: [dictEntry objectAtIndex: i]];
-    }
+    [files addObjectsFromArray: dictEntry];
   }		
   
   procfiles = [NSMutableArray new];
@@ -706,7 +739,14 @@ static NSString *nibName = @"FileOperationWin";
 	int i;
     
 	samename = NO;
-	
+
+  if (([operation isEqual: @"GWorkspaceRenameOperation"])
+        || ([operation isEqual: @"GWorkspaceCreateDirOperation"])
+        || ([operation isEqual: @"GWorkspaceCreateFileOperation"])) {
+    /* already checked by GWorkspace */
+	  return NO;
+  }
+  
 	if (destination && [files count]) {
 		dirContents = [fm directoryContentsAtPath: destination];
 		for (i = 0; i < [files count]; i++) {
@@ -725,14 +765,12 @@ static NSString *nibName = @"FileOperationWin";
           || ([operation isEqual: @"NSWorkspaceCopyOperation"])
           || ([operation isEqual: @"NSWorkspaceLinkOperation"])
           || ([operation isEqual: @"GWorkspaceRecycleOutOperation"])) {
-          
       return YES;
       
 		} else if (([operation isEqual: @"NSWorkspaceDestroyOperation"]) 
           || ([operation isEqual: @"NSWorkspaceDuplicateOperation"])
           || ([operation isEqual: @"NSWorkspaceRecycleOperation"])
           || ([operation isEqual: @"GWorkspaceEmptyRecyclerOperation"])) {
-			
       return NO;
 		} 
 	}
@@ -747,48 +785,69 @@ static NSString *nibName = @"FileOperationWin";
 
 - (oneway void)calculateNumFiles
 {
-	BOOL isDir;
-  NSDirectoryEnumerator *enumerator;
-  NSString *dirEntry;
-  int i;
+  int i, fnum = 0;
 
   for (i = 0; i < [files count]; i++) {
+    CREATE_AUTORELEASE_POOL (arp);
     NSDictionary *dict = [files objectAtIndex: i];
     NSString *name = [dict objectForKey: @"name"]; 
     NSString *path = [source stringByAppendingPathComponent: name];       
-
-	  isDir = NO;
+	  BOOL isDir = NO;
+    
 	  [fm fileExistsAtPath: path isDirectory: &isDir];
+    
 	  if (isDir) {
-      enumerator = [fm enumeratorAtPath: path];
-      while ((dirEntry = [enumerator nextObject])) {
-        if (stopped) {
+      NSDirectoryEnumerator *enumerator = [fm enumeratorAtPath: path];
+      
+      while (1) {
+        CREATE_AUTORELEASE_POOL (arp2);
+        NSString *dirEntry = [enumerator nextObject];
+      
+        if (dirEntry) {
+          if (stopped) {
+            break;
+          }
+			    fnum++;
+        
+        } else {
+          RELEASE (arp2);
           break;
         }
-			  fcount++;
+      
+        RELEASE (arp2);
       }
+            
 	  } else {
-		  fcount++;
+		  fnum++;
 	  }
     
     if (stopped) {
+      RELEASE (arp);
       break;
     }
+    
+    RELEASE (arp);
   }
 
   if (stopped) {
     [self done];
   }
 
-  [fileOp setNumFiles: fcount];
+  fcount = 0;
+  stepcount = 0;
+  
+  if (fnum < PROGR_STEPS) {
+    progstep = 1.0;
+  } else {
+    progstep = fnum / PROGR_STEPS;
+  }
+  
+  [fileOp setNumFiles: fnum];
 }
 
 - (oneway void)performOperation
 {
 	canupdate = YES; 
-  stopped = NO;
-  paused = NO;
-	fcount = [files count];
           
 	if ([operation isEqual: @"NSWorkspaceMoveOperation"]
 						|| [operation isEqual: @"GWorkspaceRecycleOutOperation"]) {
@@ -804,6 +863,12 @@ static NSString *nibName = @"FileOperationWin";
 		[self doDuplicate];
 	} else if ([operation isEqual: @"NSWorkspaceRecycleOperation"]) {
 		[self doTrash];
+	} else if ([operation isEqual: @"GWorkspaceRenameOperation"]) {
+		[self doRename];
+	} else if ([operation isEqual: @"GWorkspaceCreateDirOperation"]) {
+		[self doNewFolder];
+	} else if ([operation isEqual: @"GWorkspaceCreateFileOperation"]) {
+		[self doNewFile];
   }
 }
 
@@ -817,22 +882,38 @@ if (([files count] == 0) || stopped || paused) break
 
 #define GET_FILENAME \
 fileinfo = [files objectAtIndex: 0]; \
-filename =  [fileinfo objectForKey: @"name"]
+RETAIN (fileinfo); \
+filename = [fileinfo objectForKey: @"name"]; 
 
 - (void)doMove
 {
   while (1) {
+	  CHECK_DONE;	
 	  GET_FILENAME;    
 
     if ((samename == NO) || (samename && [self removeExisting: fileinfo])) {
-	    [fm movePath: [source stringByAppendingPathComponent: filename]
-				    toPath: [destination stringByAppendingPathComponent: filename]
-	 		     handler: self];
-      [procfiles addObject: filename];	
+      NSString *src = [source stringByAppendingPathComponent: filename];
+      NSString *dst = [destination stringByAppendingPathComponent: filename];
+      
+	    if ([fm movePath: src toPath: dst handler: self]) {    
+        [procfiles addObject: filename];	
+      
+      } else {
+        /* check for broken symlink */
+        NSDictionary *attributes = [fm fileAttributesAtPath: src traverseLink: NO];
+        
+        if (attributes && ([attributes fileType] == NSFileTypeSymbolicLink)
+                                        && ([fm fileExistsAtPath: src] == NO)) {
+          if ([fm copyPath: src toPath: dst handler: self]
+                          && [fm removeFileAtPath: src handler: self]) {
+        	  [procfiles addObject: filename];
+          }
+        }
+      }
     }
-	  [files removeObject: fileinfo];	
     
-	  CHECK_DONE;	
+ 	  [files removeObject: fileinfo];	
+    RELEASE (fileinfo);
   }
   
   if (([files count] == 0) || stopped) {
@@ -843,17 +924,18 @@ filename =  [fileinfo objectForKey: @"name"]
 - (void)doCopy
 {
   while (1) {
+	  CHECK_DONE;	
 	  GET_FILENAME;   
 
     if ((samename == NO) || (samename && [self removeExisting: fileinfo])) {
-	    [fm copyPath: [source stringByAppendingPathComponent: filename]
-				    toPath: [destination stringByAppendingPathComponent: filename]
-	 		    handler: self];
-      [procfiles addObject: filename];	     
+	    if ([fm copyPath: [source stringByAppendingPathComponent: filename]
+				        toPath: [destination stringByAppendingPathComponent: filename]
+	 		         handler: self]) {
+        [procfiles addObject: filename];	
+      }
     }
-	  [files removeObject: fileinfo];	 
-    
-	  CHECK_DONE;	
+	  [files removeObject: fileinfo];	
+    RELEASE (fileinfo); 
   }
 
   if (([files count] == 0) || stopped) {
@@ -864,17 +946,19 @@ filename =  [fileinfo objectForKey: @"name"]
 - (void)doLink
 {
   while (1) {
+	  CHECK_DONE;	
 	  GET_FILENAME;    
     
     if ((samename == NO) || (samename && [self removeExisting: fileinfo])) {
-	    [fm linkPath: [source stringByAppendingPathComponent: filename]
-				    toPath: [destination stringByAppendingPathComponent: filename]
-	 	      handler: self];
-      [procfiles addObject: filename];	      
+      NSString *dst = [destination stringByAppendingPathComponent: filename];
+      NSString *src = [source stringByAppendingPathComponent: filename];
+  
+      if ([fm createSymbolicLinkAtPath: dst pathContent: src]) {
+        [procfiles addObject: filename];	      
+      }
     }
-	  [files removeObject: fileinfo];	    
-    
-	  CHECK_DONE;	
+	  [files removeObject: fileinfo];	   
+    RELEASE (fileinfo);     
   }
 
   if (([files count] == 0) || stopped) {
@@ -885,15 +969,15 @@ filename =  [fileinfo objectForKey: @"name"]
 - (void)doRemove
 {
   while (1) {
-	  GET_FILENAME;  
-	
-	  [fm removeFileAtPath: [destination stringByAppendingPathComponent: filename]
-				         handler: self];
-
-    [procfiles addObject: filename];	 
-	  [files removeObject: fileinfo];	
-    
 	  CHECK_DONE;	
+	  GET_FILENAME;  
+	  
+	  if ([fm removeFileAtPath: [destination stringByAppendingPathComponent: filename]
+				             handler: self]) {
+      [procfiles addObject: filename];
+    }
+	  [files removeObject: fileinfo];	 
+    RELEASE (fileinfo);   
   }
 
   if (([files count] == 0) || stopped) {
@@ -903,24 +987,34 @@ filename =  [fileinfo objectForKey: @"name"]
 
 - (void)doDuplicate
 {
-  NSString *copystr = NSLocalizedString(@"copy", @"");
-  NSString *ofstr = NSLocalizedString(@"_of_", @"");
+  NSString *copystr = NSLocalizedString(@"_copy", @"");
+  NSString *base;
+  NSString *ext;
 	NSString *destpath;
 	NSString *newname;
   NSString *ntmp;
 
   while (1) {
     int count = 1;
-    
+
+	  CHECK_DONE;    
 	  GET_FILENAME;  
 
 	  newname = [NSString stringWithString: filename];
-
+    ext = [newname pathExtension]; 
+    base = [newname stringByDeletingPathExtension];
+    
 	  while (1) {
       if (count == 1) {
-        ntmp = [NSString stringWithFormat: @"%@%@%@", copystr, ofstr, newname];
+        ntmp = [NSString stringWithFormat: @"%@%@", base, copystr];
+        if ([ext length]) {
+          ntmp = [ntmp stringByAppendingPathExtension: ext];
+        }
       } else {
-        ntmp = [NSString stringWithFormat: @"%@%i%@%@", copystr, count, ofstr, newname];
+        ntmp = [NSString stringWithFormat: @"%@%@%i", base, copystr, count];
+        if ([ext length]) {
+          ntmp = [ntmp stringByAppendingPathExtension: ext];
+        }
       }
       
 		  destpath = [destination stringByAppendingPathComponent: ntmp];
@@ -933,14 +1027,13 @@ filename =  [fileinfo objectForKey: @"name"]
       }
 	  }
 
-	  [fm copyPath: [destination stringByAppendingPathComponent: filename]
-				  toPath: destpath 
-			   handler: self];
-
-    [procfiles addObject: newname];	 
-	  [files removeObject: fileinfo];	   
-    
-	  CHECK_DONE;
+	  if ([fm copyPath: [destination stringByAppendingPathComponent: filename]
+				      toPath: destpath 
+			       handler: self]) {
+      [procfiles addObject: newname];	
+    }
+	  [files removeObject: fileinfo];
+    RELEASE (fileinfo);	       
   }
   
   if (([files count] == 0) || stopped) {
@@ -948,29 +1041,95 @@ filename =  [fileinfo objectForKey: @"name"]
   }                                             
 }
 
+- (void)doRename
+{
+	GET_FILENAME;    
+  
+	if ([fm movePath: source toPath: destination handler: self]) {         
+    [procfiles addObject: filename];
+  
+  } else {
+    /* check for broken symlink */
+    NSDictionary *attributes = [fm fileAttributesAtPath: source traverseLink: NO];
+  
+    if (attributes && ([attributes fileType] == NSFileTypeSymbolicLink)
+                                    && ([fm fileExistsAtPath: source] == NO)) {
+      if ([fm copyPath: source toPath: destination handler: self]
+                      && [fm removeFileAtPath: source handler: self]) {
+        [procfiles addObject: filename];
+      }
+    }
+  }
+  
+	[files removeObject: fileinfo];
+  RELEASE (fileinfo);	
+
+  [self done];
+}
+
+- (void)doNewFolder
+{
+	GET_FILENAME;  
+
+	if ([fm createDirectoryAtPath: [destination stringByAppendingPathComponent: filename]
+				             attributes: nil]) {
+    [procfiles addObject: filename];
+  }
+	[files removeObject: fileinfo];	
+  RELEASE (fileinfo);
+
+  [self done];
+}
+
+- (void)doNewFile
+{
+	GET_FILENAME;  
+
+	if ([fm createFileAtPath: [destination stringByAppendingPathComponent: filename]
+				          contents: nil
+                attributes: nil]) {
+    [procfiles addObject: filename];
+  }
+	[files removeObject: fileinfo];	
+  RELEASE (fileinfo);
+  
+  [self done];
+}
+
 - (void)doTrash
 {
-  NSString *copystr = NSLocalizedString(@"copy", @"");
-  NSString *ofstr = NSLocalizedString(@"_of_", @"");
+  NSString *copystr = NSLocalizedString(@"_copy", @"");
+  NSString *srcpath;
 	NSString *destpath;
 	NSString *newname;
   NSString *ntmp;
 
-  while (1) {    
+  while (1) {
+	  CHECK_DONE;      
 	  GET_FILENAME;  
 
     newname = [NSString stringWithString: filename];
+    srcpath = [source stringByAppendingPathComponent: filename];
     destpath = [destination stringByAppendingPathComponent: newname];
     
     if ([fm fileExistsAtPath: destpath]) {
-	    newname = [NSString stringWithString: filename];
+      NSString *ext = [filename pathExtension]; 
+      NSString *base = [filename stringByDeletingPathExtension]; 
       int count = 1;
-
+      
+      newname = [NSString stringWithString: filename];
+      
 	    while (1) {
         if (count == 1) {
-          ntmp = [NSString stringWithFormat: @"%@%@%@", copystr, ofstr, newname];
+          ntmp = [NSString stringWithFormat: @"%@%@", base, copystr];
+          if ([ext length]) {
+            ntmp = [ntmp stringByAppendingPathExtension: ext];
+          }
         } else {
-          ntmp = [NSString stringWithFormat: @"%@%i%@%@", copystr, count, ofstr, newname];
+          ntmp = [NSString stringWithFormat: @"%@%@%i", base, copystr, count];
+          if ([ext length]) {
+            ntmp = [ntmp stringByAppendingPathExtension: ext];
+          }
         }
 
 		    destpath = [destination stringByAppendingPathComponent: ntmp];
@@ -984,14 +1143,24 @@ filename =  [fileinfo objectForKey: @"name"]
 	    }
     }
 
-	  [fm movePath: [source stringByAppendingPathComponent: filename]
-				  toPath: destpath 
-			   handler: self];
+	  if ([fm movePath: srcpath toPath: destpath handler: self]) {
+      [procfiles addObject: newname];	
+      
+    } else {
+      /* check for broken symlink */
+      NSDictionary *attributes = [fm fileAttributesAtPath: srcpath traverseLink: NO];
     
-    [procfiles addObject: newname];	 
-	  [files removeObject: fileinfo];	   
+      if (attributes && ([attributes fileType] == NSFileTypeSymbolicLink)
+                                  && ([fm fileExistsAtPath: srcpath] == NO)) {
+        if ([fm copyPath: srcpath toPath: destpath handler: self]
+                          && [fm removeFileAtPath: srcpath handler: self]) {
+          [procfiles addObject: newname];
+        }
+      }
+    }
     
-	  CHECK_DONE;
+	  [files removeObject: fileinfo];	 
+    RELEASE (fileinfo);  
   }
   
   if (([files count] == 0) || stopped) {
@@ -1003,10 +1172,11 @@ filename =  [fileinfo objectForKey: @"name"]
 {
   NSString *fname =  [info objectForKey: @"name"];
 	NSString *destpath = [destination stringByAppendingPathComponent: fname]; 
+  BOOL isdir;
     
 	canupdate = NO; 
   
-	if ([fm fileExistsAtPath: destpath]) {
+	if ([fm fileExistsAtPath: destpath isDirectory: &isdir]) {
     if (onlyolder) {
       NSDictionary *attributes = [fm fileAttributesAtPath: destpath traverseLink: NO];
       NSDate *dstdate = [attributes objectForKey: NSFileModificationDate];
@@ -1031,19 +1201,19 @@ filename =  [fileinfo objectForKey: @"name"]
   return YES;
 }
 
-- (void)Pause
+- (NSDictionary *)infoForFilename:(NSString *)name
 {
-  paused = YES;
-}
+  int i;
 
-- (void)Stop
-{
-  stopped = YES;
-}
+  for (i = 0; i < [files count]; i++) {
+    NSDictionary *info = [files objectAtIndex: i];
 
-- (BOOL)isPaused
-{
-  return paused;
+    if ([[info objectForKey: @"name"] isEqual: name]) {
+      return info;
+    }
+  }
+  
+  return nil;
 }
 
 - (void)done
@@ -1052,74 +1222,93 @@ filename =  [fileinfo objectForKey: @"name"]
   [fileOp endOperation];  
 }
 
-- (oneway void)exitThread
-{
-  [NSThread exit];
-}
-
 - (BOOL)fileManager:(NSFileManager *)manager 
               shouldProceedAfterError:(NSDictionary *)errorDict
 {  
-  NSString *path, *msg;
-  BOOL iserror = NO;
+  NSString *path;
+  NSString *error;
+  NSString *msg;
   int result;
-  
-  path = [errorDict objectForKey: @"Path"];
+
+  error = [errorDict objectForKey: @"Error"];
+
+  if ([error hasPrefix: @"Unable to change NSFileOwnerAccountID to to"]
+        || [error hasPrefix: @"Unable to change NSFileOwnerAccountName to"]
+        || [error hasPrefix: @"Unable to change NSFileGroupOwnerAccountID to"]
+        || [error hasPrefix: @"Unable to change NSFileGroupOwnerAccountName to"]
+        || [error hasPrefix: @"Unable to change NSFilePosixPermissions to"]
+        || [error hasPrefix: @"Unable to change NSFileModificationDate to"]) {
+    return YES;
+  }
+
+  path = [NSString stringWithString: [errorDict objectForKey: @"Path"]];
   
   msg = [NSString stringWithFormat: @"%@ %@\n%@ %@\n",
 							NSLocalizedString(@"File operation error:", @""),
-							[errorDict objectForKey: @"Error"],
+							error,
 							NSLocalizedString(@"with file:", @""),
 							path];
 
   result = [fileOp requestUserConfirmationWithMessage: msg title: @"Error"];
     
-	if(result != NSAlertDefaultReturn) {
-    [fileOp sendDidChangeNotification];
-    [fileOp endOperation];
+  if (result != NSAlertDefaultReturn) {
+    [self done];
     
 	} else {  
-    NSString *fname = [path lastPathComponent];
     BOOL found = NO;
     
-    while (1) {     
-      if ([path isEqualToString: source] == YES) {
+    while (1) { 
+      NSDictionary *info = [self infoForFilename: [path lastPathComponent]];
+          
+      if ([path isEqual: source]) {
         break;      
       }    
      
-      if ([files containsObject: fname] == YES) {
-        [files removeObject: fname];
+      if (info) {
+        [files removeObject: info];
         found = YES;
         break;
       }
          
       path = [path stringByDeletingLastPathComponent];
-      fname = [path lastPathComponent];
     }   
     
-    if (found == YES) {
-      [self performOperation]; 
+    if ([files count]) {
+      if (found) {
+        [self performOperation]; 
+      } else {
+        result = [fileOp showErrorAlertWithMessage: @"File Operation Error!"];
+        [self done];
+      }
     } else {
-      result = [fileOp showErrorAlertWithMessage: @"File Operation Error!"];
-      [fileOp sendDidChangeNotification];
-      [fileOp endOperation];
-      return NO;
+      [self done];
     }
   }
   
-	return !iserror;
+	return YES;
 }
 
 - (void)fileManager:(NSFileManager *)manager willProcessPath:(NSString *)path
 {
   if (canupdate) {
-    [fileOp updateProgressIndicator];
+    fcount++;
+    stepcount++;
+    
+    if (stepcount >= progstep) {
+      stepcount = 0;
+      [fileOp setProgIndicatorValue: fcount];
+    }
   }
+  
+  if (stopped) {
+    [self done];
+  }                                             
 }
 
 @end
 
-@implementation ProgressView
+
+@implementation OpProgressView
 
 #define PROG_IND_MAX (-28)
 
@@ -1135,7 +1324,10 @@ filename =  [fileinfo objectForKey: @"name"]
   self = [super initWithFrame: frameRect];
 
   if (self) {
-    ASSIGN (image, [NSImage imageNamed: @"progind.tiff"]);
+    NSBundle *bundle = [NSBundle bundleForClass: [Operation class]];
+    NSString *path = [bundle pathForResource: @"progind" ofType: @"tiff"];
+ 
+    image = [[NSImage alloc] initWithContentsOfFile: path];     
     rfsh = refresh;
     orx = PROG_IND_MAX;
   }
